
#include<iostream>
#include<vector>
using namespace std;

class SudokuSolver
{
    private:
        
        int n;
        int m;
        vector<vector<int> > board;
        vector<vector<bool> > lines;
        vector<vector<bool> > columns;
        vector<vector<bool> > blocks;
        int recursions;
        bool allsolutions;
        int solutioncount;

    public:
       
        SudokuSolver(int chunksize, bool all_solutions)
        {
            if(chunksize < 1)
                chunksize = 3;
            m = chunksize;
            n = m*m;
            recursions=0;
            solutioncount=0;
            allsolutions = all_solutions;
            board = vector<vector<int> >(n, vector<int>(n, 0));
            lines = vector<vector<bool> >(n, vector<bool>(n, true));
            columns = vector<vector<bool> >(n, vector<bool>(n, true));
            blocks = vector<vector<bool> >(n, vector<bool>(n, true));
        }
         
        ~SudokuSolver()
        {
        }

        void print()
        {
            for(int i = 0; i < n; i++)
            {
                cout << board[0][i];
                for(int j = 1; j < n; j++)
                {
                    cout << " " << board[j][i];
                    if((j+1)%m == 0)
                        cout << " ";
                }
                cout << "\n";
                if((i+1)%m == 0)
                    cout << "\n";
            }
            cout << "\n";
        }


        int read()
        {
            int defined = 0;
            int value;
            for(int y = 0; y < n; y++)
            {
                for(int x = 0; x < n; x++)
                {
                    cin >> value;
                    if(value <= 0 || value > n)
                        board[x][y] = 0;
                    else
                    {
                        put(x+1,y+1,value);
                        defined++;
                    }
                }
            }
            return defined;
        }
         
        int GetRecursions()
        {
            return recursions;
        }
        
        int GetSolutionCount()
        {
            return solutioncount;
        }


        bool put(int x, int y, int number)
        {
            x--;
            y--;
            number--;
            int blocknum = (y/m)*m+(x/m);

            if(x < 0 || x >= n || y < 0 || y >= n || number < 0 || number >= n)
                return false;
            if(!lines[y][number] || !columns[x][number] || !blocks[blocknum][number])
                return false;

            board[x][y] = number+1;
            lines[y][number] = false;
            columns[x][number] = false;
            blocks[blocknum][number] = false;
        }

        bool remove(int x, int y)
        {
           x--;
           y--;
           int number = board[x][y]-1;
           if(number == -1)
               return false;
           int blocknum = (y/m)*m+(x/m);

           if(x < 0 || x >= n || y < 0 || y >= n || number < 0 || number > n)
               return false;
           if(lines[y][number] || columns[x][number] || blocks[blocknum][number])
               return false;

           board[x][y] = 0;
           lines[y][number] = true;
           columns[x][number] = true;
           blocks[blocknum][number] = true;
        }

        bool solve(int rest, int depth = 0, bool log = false)
        {
            if(rest-- <= 0)
            {
                if(log)
                {
                    for(int i = 0; i < depth; i++)
                        cerr << "   ";
                    cerr << "Success!\n";
                }
                print();
                solutioncount++;
                return true;
            }

            ++recursions;
            int CurrentCandidateCount;
            int BestCandidateCount = n+1;
            int BestCandidateX;
            int BestCandidateY;
            int BestCandidateBlocknum;

            // find cell with minimal branching factor
            int blocknum;
            for(int y = 0; y < n; y++)
                for(int x = 0; x < n; x++)
                    if(board[x][y] == 0)
                    {
                        blocknum = (y/m)*m+(x/m);
                        CurrentCandidateCount = 0;
                        for(int candidate = 0; candidate < n; candidate++)
                            if(lines[y][candidate] && columns[x][candidate] && blocks[blocknum][candidate])
                                if(++CurrentCandidateCount > BestCandidateCount)
                                    break;

                        if(CurrentCandidateCount < BestCandidateCount)
                        {
                            BestCandidateCount = CurrentCandidateCount;
                            BestCandidateX = x;
                            BestCandidateY = y;
                            BestCandidateBlocknum = blocknum;
                        }
                    }

            // write log
            if(log)
            {
                for(int i = 0; i < depth; i++)
                    cerr << "   ";
                if(BestCandidateCount < 1)
                    cerr << "No Candidates for ("<<(BestCandidateX+1)<<","<<(BestCandidateY+1)<<")\n";
                else
                {
                    cerr << "Best Cell: ("<<(BestCandidateX+1)<<","<<(BestCandidateY+1)<<"), branching factor: " << BestCandidateCount << ", candidates: ";
                    for(int candidate = 0; candidate < n; candidate++)
                        if(lines[BestCandidateY][candidate] && columns[BestCandidateX][candidate] && blocks[BestCandidateBlocknum][candidate])
                            cerr << (candidate+1) << " ";
                    cerr << "\n";
                }
            }

            // backtrack
            for(int candidate = 0; candidate < n; candidate++)
                if(lines[BestCandidateY][candidate] && columns[BestCandidateX][candidate] && blocks[BestCandidateBlocknum][candidate])
                {
                    if(log)
                    {
                        for(int i = 0; i < depth; i++)
                            cerr << "   ";
                        cerr << "Trying Candidate " << (candidate+1) << "\n";
                    }

                    put(BestCandidateX+1, BestCandidateY+1, candidate+1);
                    if(solve(rest, depth+1))
                        if(!allsolutions)
                            return true;
                    remove(BestCandidateX+1, BestCandidateY+1);
                }
                
            return false;
        }
};





int main(int args, char** argv)
{
    bool allsolutions = true;
    SudokuSolver S(3, allsolutions);
    int defined = S.read();
    if(defined >= 81)
    {
        S.print();
        cerr << "nothing to do\n";
        return 0;
    }

    cout << "\n";
    if(allsolutions)
    {
        S.solve(81-defined);
        cerr << S.GetSolutionCount() << " solution" << (S.GetSolutionCount()>1?"s":"") << " found\n";
        cerr << S.GetRecursions() << " recursions needed\n";
    }
    else
    {
        if(!S.solve(81-defined))
        {
            cerr << "no solution found\n";
            cerr << S.GetRecursions() << " recursions needed\n";
            return 1;
        }
        S.print();
        cerr << S.GetRecursions() << " recursions needed\n";
    }
    
    return 0;
}

